"""Server connection package.

The package handles connecting to the server, as well as all the related
issues. It also starts a sending and receiving thread with the resoective
sending/receiving queue.
"""
